#!/usr/bin/env python3

from ..warn.warning import *
from ..err.error import *

import matplotlib.colors as mcolors
from matplotlib import markers
import re


"""
Handle all the input messages in EMMER
"""
class TestInputLength:
    """
    Make sure that the number of element in input matches the number of cluster, group,
    or individuals.
    """
    def __init__(self, set, input, expect = 0):
        self.set = set
        self.input = input
        self.expect = expect

    def testLen(self):
        try:
            if len(self.set) != len(self.input):
                raise ErrorCode10(add_msg = [self.input, self.set], suppress = False, second_chance = True)
            else:
                self.dictionary = dict(zip(self.set, self.input))

        except ErrorCode10:
            new_input = input('Please re-enter your input:\n').split(';')  ## TODO: not protected

            try:
                if len(self.set) != len(new_input):
                     raise ErrorCode10(add_msg = [self.input, self.set], suppress = False, second_chance = False)
                else:
                     self.input = new_input
                     self.dictionary = dict(zip(self.set, self.input))  #zip(keys, values)
            except ErrorCode10:
                pass

    # TODO: check input type


class InputCode1:
    """
    Input edge or fill color:
    Please assign color to each group or individuals.
    Please use semicolon to separate the assignment for each group or individuals.

    Location:
        individual.py

    Attributes:
        color_dict -- Type: dictionary
                      key: group; item: color
    """

    def __init__(self, set, add_msg, suppress, second_chance):
        count = 0
        colors = input(f'\n[[Input code 1]] Input {add_msg[0]} color:\nPlease assign color to each {add_msg[1]}.\n Please use semicolon to separate the assignment for each {add_msg[1]}\nExample 1:\nred;blue\nExample 2:\n#ff0000;#0000ff\n').split(';')
        check = TestInputLength(set = set, input = colors)
        check.testLen()

        ## first entry
        for element in check.input:
            if element in mcolors.cnames.keys() or element == 'none':
                pass
            else:
                hex_color = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', element)
                # ref: https://stackoverflow.com/questions/30241375/python-how-to-check-if-string-is-a-hex-color-code

                try:
                    if not hex_color:
                        raise ErrorCode28(add_msg = add_msg, suppress = suppress, second_chance = second_chance)
                except ErrorCode28:

        ## second entry
                    colors = input(f'\n[[Input code 1]] Input {add_msg[0]} color:\nPlease assign color to each {add_msg[1]}.\n Please use semicolon to separate the assignment for each {add_msg[1]}\nExample 1:\nred;blue\nExample 2:\n#ff0000,#0000ff\n').split(';')
                    check = TestInputLength(set = set, input = colors)
                    check.testLen()
                    print(check.input)

                    for element in check.input:
                        if element in mcolors.cnames.keys() or element == 'none':
                            pass
                        else:
                            hex_color = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', element)
                            # ref: https://stackoverflow.com/questions/30241375/python-how-to-check-if-string-is-a-hex-color-code

                            try:
                                if not hex_color:
                                    raise ErrorCode28(add_msg = add_msg, suppress = suppress, second_chance = False)

                            except ErrorCode28:
                                pass

        ## pass all tests, save in attribute
        self.color_dict = check.dictionary


class InputCode2:
    """
    Input cluster assignment:
    Please assign cluster to each group.

    Location:
        permanova.py

    Attributes:
        color_dict -- Type: dictionary
                      key: group; item: color
    """

    def __init__(self, set, suppress, second_chance):
        count = 0
        cluster = input(f'\n[[Input code 2]] Input cluster assignment:\nPlease assign cluster to each group.\n').split(';')
        check = TestInputLength(set = set, input = cluster)
        check.testLen()

        ## first entry
        try:
            if len(check.dictionary) > 2:
                raise ErrorCode37(suppress = suppress, second_chance = True)

        ## second entry
        except ErrorCode37:
            cluster = input(f'\n[[Input code 2]] Input cluster assignment:\nPlease assign cluster to each group.\n').split(';')
            check = TestInputLength(set = set, input = cluster)
            check.testLen()

            try:
                if len(check.dictionary) < 2:
                    raise ErrorCode37(suppress = suppress, second_chance = False)
            except ErrorCode37:
                pass

        ## pass all tests, save in attribute
        print(check.dictionary)
        self.cluster_dict = check.dictionary


class InputCode3:
    """
    Multiple choices:
    Ask user to choice one from the list.

    Location:
        bake.py

    Attributes:
        choice --
    """

    def __init__(self, choice, add_msg, suppress, second_chance):
        option = input(f'[[Input code 3]] {add_msg}. Acceptable choice includes: {choice}\n')
        # add_msg = 'please choose retrospect mode'
        # choice = ['Individual', 'Permanova', 'Reproducibility', 'RevisitThreshold', 'Bifurication']

        ## first entry
        try:
            if option not in choice:
                raise ErrorCode9(add_msg = choice, suppress = False, second_chance = True)

        ## second entry
        except ErrorCode9:
            option = input(f'[[Input code 3]] {add_msg}. Acceptable choice includes: {choice}\n')

            try:
                if option not in choice:
                    raise ErrorCode9(add_msg = choice, suppress = False, second_chance = False)
            except ErrorCode9:
                pass

        self.select = option


class InputCode4:
    """
    Please assign shape to each group or individuals.
    Please semicolon comma to separate the assignment for each group or individuals.
    More shape information (https://matplotlib.org/api/markers_api.html).

    Location:
        individual.py

    Attributes:
        color_dict -- Type: dictionary
                      key: group; item: shape
    """

    def __init__(self, set, add_msg, suppress, second_chance):
        count = 0
        shape = input(f'[[Input code 4]] Please assign shape style to each {add_msg}.\nPlease use semicolon to separate the assignment for each {add_msg}.\nExample:\no;s\nMore shape information (https://matplotlib.org/api/markers_api.html).\n').split(';')
        check = TestInputLength(set = set, input = shape)
        check.testLen()

        ## first entry
        for element in check.input:
            try:
                if element not in markers.MarkerStyle.markers.keys():
                    raise ErrorCode33(add_msg = add_msg, suppress = suppress, second_chance = second_chance)

        ## second entry
            except ErrorCode33:
                shape = input(f'[[Input code 4]] Please assign shape style to each {add_msg}.\nPlease use semicolon to separate the assignment for each {add_msg}.\nExample:\no;s\nMore shape information (https://matplotlib.org/api/markers_api.html).\n').split(';')
                check = TestInputLength(set = set, input = shape)
                check.testLen()

                try:
                    for element in check.input:
                        if element in markers.MarkerStyle.markers.keys():
                            pass
                        else:
                            raise ErrorCode33(add_msg = add_msg, suppress = suppress, second_chance = False)
                except ErrorCode33:
                    pass

        ## pass all tests, save in attribute
        self.shape_dict = check.dictionary


class InputCode5:
    """
    Allow users to choose whether they want to color by group or by individuals

    Location:
        individual.py

    Attributes:
        decision -- Type: string
    """

    def __init__(self, add_msg, suppress, second_chance):
        decision = input(f'\n[[Input Code 5]] Please decide whether you want to set {add_msg} by group or individual.\nPlease enter one of following options ["Group", "Individaul"]\n')

        try:
            if decision in ["Group", "Individaul"]:
                self.decision = decision
            else:
                raise ErrorCode36(add_msg = add_msg, suppress = suppress, second_chance = second_chance)
        except ErrorCode36:
            decision = input(f'\n[[Input Code 5]] Please decide whether you want to set {add_msg} by group or individual.\nPlease enter one of following options ["Group", "Individaul"]\n')

            try:
                if decision in ["Group", "Individaul"]:
                    self.decision = decision
                else:
                    raise ErrorCode36(add_msg = add_msg, suppress = suppress, second_chance = False)
            except ErrorCode36:
                pass



#!/usr/bin/env python3

# Usage:
# move to one level above emmer/
# python3 -m emmer.harvest -i emmer/data/data_dir_3 -f 'HardFilter' -u 2 -l 2 -t 2 -d 0.001 -z 0.33 -r

from .main.basic.math import NonDesityMatrix
from .main.basic.read import GetFiles, RawDataImport
#from .main.advanced.iteration import MinusOneVNE, MinDataLostFilter, InfoRichCalling, reproducibility, reproducibility_summary, Kernal
from .main.advanced.iteration import MinusOneVNE, InfoRichCalling, reproducibility, reproducibility_summary, Kernal
from .posthoc.visual.viewer import Projection, Plot
from .troubleshoot.warn.warning import *
from .troubleshoot.err.error import *
from .toolbox.recorder import initNoteBook, UpdateNoteBook
from .toolbox.technical import emptyNumpyArray

from scipy.spatial import procrustes
import itertools
import argparse
import pandas
import numpy
import math
import time
import glob
import sys
import os

start_time = time.time()
__version__ = '1.0'

"""
Take user-defined args and run emmer
"""

##==1==## define function and class
def tutorial():
    """
    ##==1==## Expected Data Format
    1. Expected csv file(s). When input multiple csv files as input, those csv files
       needed to be stored under the same folder. That folder should only contains your
       input csv file(s).
    2. Expect to have row names and column header. Examplary input file can be found
       under 'emmer/data/data_dir_3/'
       row names: sample names
       column headers: feature names
       elements: numeric values with on missing value
                 (Current version of emmer does not support input matrix with missing
                 values. In future update, the missing value will be converted to 0)
    3. Do not include '__' in your file name or row names

    ##==2==## Output Files and Format
    1. output/*_reproducibility.csv
       Store informaiton-rich feature calling reproducibilty. Emmer will not generate
       this file when running under QuickLook mode,
    2. output/information_rich_features_summary.csv
       Generate a summary when emmer working on multiple input files.
    3. output/*_colmean.csv
       Store means for each feature. Necessary for projecting new observation onto
       the existing PCA space.
    4. output/*_transformation_matrix.csv
       Store left signular matrix. Necessary for projecting new observation onto
       the existing PCA space.
    5. output/*_coordinates.csv or *_projection.csv
       Coordinates on the PCA space
    6. *_PCA_*.pdf
       PCA plot
    [!] The name and number of output files might varys depend on user's settings. Please
        refer to the Arugments setions for detail information about settings

    ##==3==## Quick Examples
    1. Input matrix contains counts, but you wish to converted the counts into fractional
       abundance before conducting your analysis.
       python3 -m emmer.harvest -i emmer/data/data_dir_3 -f 'HardFilter' -u 2 -l 2 -t 2 -d 0.001 -z 0.33 -r

    2. If you input data contains categorical values and want to conduct PCA by eigendecomposition
       of a correlation matrix.
       (Current version of emmer does not support this function. Will included in future update)

    ##==4==## Suggested Analytical Flow
    [Time is of the essence]
    1. A quick look
       python3 -m emmer.harvest -i emmer/data/data_dir_3 -f 'HardFilter' -u 2 -l 2 -d 0.001 -z 0.33 -r -p -q

    [Just what to be thorough]
    2. Report the reproducibilty of information-rich feature calling
       python3 -m emmer.harvest -i emmer/data/data_dir_3 -f 'HardFilter' -u 2 -l 2 -t 2 -d 0.001 -z 0.33 -r
    3. Fine tuning your thresholds
       python3 -m emmer.bake -m 'RevisitThreshold' -u 2.5,1.5,0.25 -l 2.5,1.5,0.25 -t 2,2,0 -e output/detail_vNE/ -i output/filtered_data/
    4. Redo analysis with the suggested information-rich feature calling thresholds

    [Picture paints a thousand words/Significant p value made my day]
    5. Export emmer.bake for additional options for generating plots and conducting statistical tests
       python3 -m emmer.bake -g

    ##==5==## Arugments
    Required:
    -i I               Input directory that contains one or many csv files or a path to specific csv file.
    -f F, -filter      Filter data before nominating information-rich features.
                       1. HardFilter: Remove column that contains -z levels of zeros. Require to set arugment -z.
                       2. None: No filer. Might take longer time to nominate information-rich feature.
    -u U               Upper limit when choosing information-rich feature. Default: 1. When set as 1, the
                       upper limitation is 1 standard deviation from the mean of the von Neumann entropies
                       of all remove-one-column matrices.
    -l L               Lower limit when choosing information-rich feature. Default: 1. When set as 1, the
                       lower limitation is 1 standard deviation from the mean of the von Neumann entropies
                       of all remove-one-column matrices.
    -t T               How many times a feature need to be nominate as information-rich feature in jackknife
                       subsampling to be included in the final list of information-rich features.

    Conditionally requried:
    -z, Z, zeroToleranceLevel
                       Must be a less than 1 but greater than 0. The number represents the maximum
                       fraction of element in each column of your input matrix that can be zero. Column that
                       exceeds this threshold will be remove before information-rich feature calling.

    Optional:
    -d D, -detectionLimit
                       Set the detection limit of our method. Must be numeric. Set number in the input
                       matrix that below this detection limit as zero before filtering. When using with -z,
                       emmer will conduct this step before running -z.
    -q Q, -quickLook   Default: False. If True, activate QuickLook mode. In QuickLook mode, emmer will not
                       calculate the reproducibility of information-rich feature calling. Default: False.
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -q
    -o O               Expect a string. This string will be tag to all your output file name from the analysis.
    -r R               Choose this option when you want emmer to convert your input data into fractional abundance.
                       Default: False.
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -r
    -n N, normalize    Choose this option when your features (columns) are measured in different units. When set as True,
                       EMMER will normalize each column the input data by its standard deviation.
                       Default: False.
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -n
    -p P, plot         Visualize emmer result with PCA plot(s). Default: False.
                       (Currently emmer cannot generate PCA plot if it only works on input file. Also,
                       user might experience error when generateing 2D PCA plots. These issue will be
                       addressed in future updates)
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -p
                       [!] An interactive plot that allows user to adjust angles. Save the plot as pdf
                           after user close the window.
    -s S, sanityCheck  Generate three or six PCA plots (and their corresponding output csv files) depend on
                       the filter of choose. Default: False.
                       1: Full dataset
                       2: filtered-out data (if filter is not set at 'None')
                       3: filtered data (if filter is not set at 'None')
                       4: filtered, non-information-rich data (if filter is not set at 'None')
                          non-information-rich data (if filter is set at 'None')
                       5: Information-rich data
                       6. filtered-out data combined with filtered, non-information-rich data
                          (if filter is not set at 'None')
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -s
    -c C, -cpuNum      Support multiprocessing. This argument is used to set the number of CPU used in the analysis.
    -w W, -writeDownDetails
                       Do you want to add additional notes as emmer run? Default: False.
                       (Not available,. Will be included in future update)
                       Usage:
                       python3 -m emmer.harvest <other_arguments_and_inputs> -w
    """
    pass


class HarvestArgs:
    """
    Take arguments and report error messages when necessary

    Objective: so we can test and use @

    Argument:
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean
                   When set as True. Silence warning messages
        neglect -- Type: boolean
                   Suppress notebook initation when running unittest

    Attributes:
        suppress -- Type: boolean
        silence -- Type: boolean
        neglect -- Type: boolean
        notebook_name -- Type: str
                         Notebook file name with full path
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        input_dir -- Type: str
                     Corresponding to args.i
        input -- Type: emmer.main.basic.read.GetFiles
        specific_csv -- Type: boolean
                        Set True if <self.input_dir> is a specific csv file. False when <self.input_dir>
                        is a directory
        quick_look -- Type: boolean
                      Corresponding to args.q
        infoRich_threshold -- Type: int
                              Corresponding to args.t
        output_file_tag -- Type: str
                           Corresponding to args.o
        use_fractional_abundance -- Type: boolean
                                    Corresponding to args.r
        detection_limit -- Type: float
                           Corresponding to args.d
        filter -- Type: str
                  Corresponding to args.f
        upper_threshold_factor -- Type: float or str ('None')
                                  Corrresponding to args.u
        lower_threshold_factor -- Type: float or str ('None')
                                  Corrresponding to args.l
        plot_result -- Type: float
                       Corresponding to args.p
        sanity_check_plots -- Type: boolean
                              Corresponding to args.s
        warning_code -- Type: str
                        For unittest
        normalize -- Type: boolean
                     Scale each column in the mean centered data based on its standard deviation
                     before SVD
    """
    def __init__(self, suppress, silence, neglect):
        parser = argparse.ArgumentParser(description = '#############################################################################\nPlease use -g when you need additional explanation on different modes their corresponding arguments. Try: python3 -m emmer.harvest -g\n#############################################################################')
        parser.add_argument('-g', '-guide', action = 'store_true')
        parser.add_argument('-i', type = str)
        parser.add_argument('-q', '-quickLook', action = 'store_true')
        parser.add_argument('-o', type = str)
        parser.add_argument('-r', action = 'store_true')
        parser.add_argument('-n', '-normalize', action = 'store_true')
        parser.add_argument('-d', '-detectionLimit', type = float)
#        parser.add_argument('-f', '-filter', type = str, choices = ['MinDataLostFilter', 'HardFilter', 'None'])
        parser.add_argument('-f', '-filter', type = str, choices = ['HardFilter', 'None'])
        parser.add_argument('-z', '-zeroToleranceLevel', type = float)
        parser.add_argument('-t', type = int)
        parser.add_argument('-u', type = float)
        parser.add_argument('-l', type = float)
        parser.add_argument('-p', '-plot', action = 'store_true')
        parser.add_argument('-s', '-sanityCheck', action = 'store_true')
        parser.add_argument('-c', '-cpuNum', default = 1, type = int)
        parser.add_argument('-w', '-writeDownDetails', action = 'store_true')
        self.args = parser.parse_args()
        self.suppress = suppress
        self.neglect = neglect
        self.silence = silence

        if self.args.g:
            print(tutorial.__doc__)
            sys.exit()


    def getArgsO(self):
        if self.args.o:
            self.output_file_tag = self.args.o
        else:
            self.output_file_tag = ''


    def getArgsW(self):
        self.notebook_name = initNoteBook(current_wd = os.getcwd(), script_name = 'emmer.harvest', script_version = __version__, tag = self.output_file_tag,
                                          neglect = self.neglect, explicit = self.args.w)


    def getArgsI(self):
        self.input_dir = self.args.i
        self.specific_csv = False
        try:
            if os.path.isdir(self.input_dir) == True:
                self.input = GetFiles(self.input_dir)
            elif self.input_dir.endswith('.csv') == True:
                self.specific_csv = True
            elif os.path.isdir(self.input_dir) == False and self.input_dir.endswith('.csv') == False:
                raise Error(code = '1')

        except Error as e:
            raise ErrorCode1(suppress = self.suppress) from e


    def getArgsQT(self):
        if self.args.q:
            self.quick_look = True
            self.infoRich_threshold = 1

            try:
                if self.args.t:
                    raise WarningCode10(silence = self.silence)
            except WarningCode10:
                self.warning_code = '10'
                self.infoRich_threshold = 1


        else:
            self.quick_look = False

            if self.args.t:
                self.infoRich_threshold = self.args.t
            else:
                self.infoRich_threshold = 2


    def getArgsR(self):
        if self.args.r:
            self.use_fractional_abundance = True
        else:
            self.use_fractional_abundance = False


    def getArgsN(self):
        if self.args.n:
            self.normalize = True
        else:
            self.normalize = False


    def getArgsD(self):
        if self.args.d:
            self.detection_limit = self.args.d
        else:
            self.detection_limit = 0


    def getArgsFZ(self):
        if self.args.f == 'HardFilter':
            self.filter = self.args.f

            # missing args.z
            try:
                if not self.args.z:
                    raise Error(code = '3')
            except Error as e:
                raise ErrorCode3(suppress = self.suppress) from e

            # incorrect args.z number setting
            try:
                if self.args.z > 0 and self.args.z <= 1:
                    self.tolerance = self.args.z
                else:
                    raise Error(code = '2')
            except Error as e:
                raise ErrorCode2(suppress = self.suppress) from e

            if self.args.d:
                self.detection_limit = self.args.d

        else:
            self.tolerance = 1
            self.filter = 'None'

            try:
                if self.args.z:
                    raise WarningCode2(silence = self.silence)
            except:
                self.warning_code = '2'


    def getArgsUL(self):
        try:
            if self.args.u and self.args.l:
                self.upper_threshold_factor = self.args.u
                self.lower_threshold_factor = self.args.l
            elif self.args.u:
                self.upper_threshold_factor = self.args.u
                self.lower_threshold_factor = 'None'
            elif self.args.l:
                self.lower_threshold_factor = self.args.l
                self.upper_threshold_factor = 'None'
            else:
                raise Error(code = '5')
        except Error as e:
            raise ErrorCode5(suppress = self.suppress) from e


    def getArgsPS(self):
        if self.args.p:
            self.plot_result = True

            if self.args.s:
                self.sanity_check_plots = True
            else:
                self.sanity_check_plots = False

        else:
            self.plot_result = False

            if self.args.s:
                self.sanity_check_plots = True
            else:
                self.sanity_check_plots = False

        # prevent user setting errors
        if self.args.s:
            try:
                if self.specific_csv == True or len(self.input.input_files) == 1:
                    raise WarningCode3(silence = self.silence)
            except:
                self.warning_code = '3'

        if self.args.p:
            try:
                if self.specific_csv == True or len(self.input.input_files) == 1:
                    raise WarningCode5(silence = self.silence)
            except:
                self.warning_code = '5'


    def getArgsC(self):
        try:
            if os.cpu_count() < self.args.c or self.args.c == 0:
                raise Error(code = '47')
            else:
                self.num_cpu = self.args.c
        except Error as e:
            raise ErrorCode47(suppress = self.suppress) from e


    def processArgs(self):
        self.getArgsO()
        self.getArgsW()
        self.getArgsI()
        self.getArgsQT()
        self.getArgsR()
        self.getArgsN()
        self.getArgsD()
        self.getArgsFZ()
        self.getArgsUL()
        self.getArgsPS()
        self.getArgsC()
        UpdateNoteBook(notebook_name = self.notebook_name, neglect = self.neglect).updateArgs(args = self.args)


class EMMER:
    """
    Take in user-defined parameters and input file(s) and identify list(s) of information-rich
    features. When ask to process multiple files, those files need to be store in the same directory.
    Also, all those files will subject to the same processing parameters.
    """

    def __init__(self, input_dir, output_file_tag, detection_limit, tolerance,
                 filter, upper_threshold_factor, lower_threshold_factor, specific_csv,
                 infoRich_threshold, num_cpu, notebook_name, neglect, quick_look,
                 use_fractional_abundance, normalize):

        self.output_file_tag = str(output_file_tag)
        self.detection_limit = detection_limit
        self.tolerance = tolerance
        self.filter = filter
        self.upper_threshold_factor = upper_threshold_factor
        self.lower_threshold_factor = lower_threshold_factor
        self.num_cpu = num_cpu
        self.infoRich_threshold = infoRich_threshold
        self.notebook_name = notebook_name
        self.neglect = neglect
        self.quick_look = quick_look
        self.use_fractional_abundance = use_fractional_abundance
        self.normalize = normalize
        self.collections_of_info_rich_features = []

        ## import all csv file store under input_dir
        if specific_csv == True:
            self.input_file_names = [input_dir]
        else:
            self.input_file_names = GetFiles(input_dir).input_files

        self.current_wd = os.getcwd()
        self.output_dir = os.path.join(self.current_wd, 'output')
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.ra_data_output_dir = os.path.join(self.output_dir, 'pre_filter_data')
        if not os.path.exists(self.ra_data_output_dir):
            os.makedirs(self.ra_data_output_dir)

        self.filter_out_data_output_dir = os.path.join(self.output_dir, 'filtered_out_data')
        if not os.path.exists(self.filter_out_data_output_dir):
            os.makedirs(self.filter_out_data_output_dir)

        self.filter_data_output_dir = os.path.join(self.output_dir, 'filtered_data')
        if not os.path.exists(self.filter_data_output_dir):
            os.makedirs(self.filter_data_output_dir)

        self.raw_not_infoRich_data_output_dir = os.path.join(self.output_dir, 'raw_not_infoRich_data')
        if not os.path.exists(self.raw_not_infoRich_data_output_dir):
            os.makedirs(self.raw_not_infoRich_data_output_dir)

        self.detail_vNE = os.path.join(self.output_dir, 'detail_vNE')
        if not os.path.exists(self.detail_vNE):
            os.makedirs(self.detail_vNE)


    def singleFile(self, current_file_no = 0, silence = False):
        self.input_basename = os.path.basename(self.input_file_names[current_file_no])
        self.silence = silence
        self.data = Kernal(file_name = self.input_file_names[current_file_no], detection_limit = self.detection_limit,
                           tolerance = self.tolerance, filter = self.filter, upper_lim = self.upper_threshold_factor,
                           lower_lim = self.lower_threshold_factor, infoRich_threshold = self.infoRich_threshold,
                           quick_look = self.quick_look, use_fractional_abundance = self.use_fractional_abundance,
                           vNE_output_folder =  self.detail_vNE, output_file_tag = self.output_file_tag, num_cpu = self.num_cpu,
                           notebook_name = self.notebook_name, normalize = self.normalize, neglect = self.neglect,
                           silence = self.silence)

        self.data.importAndProcess()

        ## modify index value of filtered dataset
        index_tag = self.input_basename.replace(".csv", "__")
        clean_df = self.data.filtered_data.data
        clean_df.index = [index_tag + element for element in self.data.filtered_data.sample_id]

        ## prepare file names
        pre_filter_data_file_name = os.path.join(self.ra_data_output_dir, self.input_basename.replace(".csv", "")) + self.output_file_tag + "__pre_filterd_data.csv"
        filter_out_data_file_name = os.path.join(self.filter_out_data_output_dir, self.input_basename.replace(".csv", "")) + self.output_file_tag + "__filterd_out_data.csv"
        clean_df_file_name = os.path.join(self.filter_data_output_dir, self.input_basename.replace(".csv", "")) + self.output_file_tag + "__filterd_data.csv"
        raw_not_infoRich_data_name = os.path.join(self.raw_not_infoRich_data_output_dir, self.input_basename.replace(".csv", "")) + self.output_file_tag + "__raw_not_infoRich_data.csv"

        if current_file_no == 0:
            self.pre_filter_data_file_names = [pre_filter_data_file_name]
            self.filter_out_data_file_names = [filter_out_data_file_name]
            self.clean_df_file_names = [clean_df_file_name]
            self.raw_not_infoRich_data_name = [raw_not_infoRich_data_name]

        else:
            self.pre_filter_data_file_names.append(pre_filter_data_file_name)
            self.filter_out_data_file_names.append(filter_out_data_file_name)
            self.clean_df_file_names.append(clean_df_file_name)
            self.raw_not_infoRich_data_name.append(raw_not_infoRich_data_name)

        clean_df.to_csv(clean_df_file_name)

        ## get prefilter data
        #   raw_data -> relativeAbundance()
        self.data.input_matrix.data.index = index_tag + self.data.input_matrix.data.index
        self.data.input_matrix.raw_data_before_filter.index = [index_tag + element for element in self.data.input_matrix.sample_id]
        self.data.input_matrix.raw_data_before_filter.to_csv(pre_filter_data_file_name)

        ## get and export filtered out dataset
        #   raw_data -> relativeAbundance(); .raw_data_before_filter !-> pass filter
        features_that_fail_at_filtering = [value for value in self.data.input_matrix.raw_data_before_filter.columns if value not in list(clean_df.columns)]
        fail_filter = self.data.input_matrix.raw_data_before_filter[features_that_fail_at_filtering]
        fail_filter.to_csv(filter_out_data_file_name)

        ## Identify information-rich features
        self.data.infoRichCallingAndReproducibility()

        if self.quick_look == False:
            self.output_file_name = os.path.join(self.output_dir, self.input_basename.replace(".csv", "")) + self.output_file_tag + "_reproducibility.csv"
            self.data.info_rich_features_w_reproducibility.to_csv(self.output_file_name)

        ## get filter_out_and_no_information-rich
        #   raw_data -> remove infoRich features
        not_infoRich_features_in_raw_data = [value for value in self.data.input_matrix.raw_data_before_filter.columns if value not in self.data.list_of_info_rich_features]
        filter_out_and_not_infoRich_data_file_name = self.data.input_matrix.raw_data_before_filter[not_infoRich_features_in_raw_data]
        filter_out_and_not_infoRich_data_file_name.to_csv(raw_not_infoRich_data_name)

        ## collect the names of information-rich features
        self.collections_of_info_rich_features.append(self.data.list_of_info_rich_features)


    def multipleFiles(self):
        ## identify information-rich feature for individaul file
        for i in range(len(self.input_file_names)):
            self.singleFile(i)

            file_name_sub = [str(self.input_basename)]
            if self.quick_look == True:
                info_rich_feature_sub = self.data.list_of_info_rich_features
                reproducibility_sub = [1] * len(info_rich_feature_sub)

            else:
                info_rich_feature_sub = self.data.list_of_info_rich_features
                reproducibility_sub = list(self.data.info_rich_features_w_reproducibility['repreducibility (%)'])

            reproducibility_sub = numpy.round(numpy.array(reproducibility_sub), decimals = 2)
            summary_df_sub = pandas.DataFrame(reproducibility_sub, columns = file_name_sub, index = info_rich_feature_sub)

            if i == 0:
                self.summary_df = summary_df_sub

            else:
                # similar to cbind() and keep all
                self.summary_df = pandas.concat([self.summary_df, summary_df_sub], axis = 1, sort = True)

        self.summary_df = self.summary_df.fillna(0)
        self.summary_df.to_csv(os.path.join(self.output_dir, "information_rich_features_summary.csv"))
        self.list_of_info_rich_features_found_in_files = list(self.summary_df.index)

        ## keep unique information-rich feature names
        self.collections_of_info_rich_features = list(set(list(itertools.chain(*self.collections_of_info_rich_features))))


def mergeDataFrame(EMMER_class, select, file_name_list, info_rich_list, notebook_name, normalize, neglect):
    """
    Allow user to merge different kinds of csv files:

    select:
    'pre_filter': prefilter csv files
    'filtered_excluded': data the fail to pass filter
    'filtered': filtered data before informaiton-rich feature calling
    'filtered_infoRich': filtered data that contains only information-rich features
    'filtered_no_infoRich': filtered data without information-rich features
    'raw_not_infoRich': raw data without information-rich features
    """

    print(f"Merging data frame with {select} as column selection condition...")

    combine_list_of_informoation_rich_features = list(EMMER_class.summary_df.index.values)

    for i in range(len(file_name_list)):
        input_basename = os.path.basename(file_name_list[i])

        input_matrix = RawDataImport(file_name = file_name_list[i], for_merging_file = True, suppress = False, second_chance = False)
        input_matrix.readCSV()

        dataframe_for_merging = pandas.DataFrame(input_matrix.data,
                                                 columns = input_matrix.feature_names,
                                                 index = input_matrix.sample_id)

        ## Need to select specific columns
        # 'filtered_infoRich', 'filtered_no_infoRich'
        #
        #  Select only the columns that were nominated as combine_list_of_informoation_rich_features
        #  Not all of the feature in self.dataframe_for_merging can be found in combine_list_of_informoation_rich_features
        #  To avoid error msg, first take the intersect between the two lists, then subset the data
        if select == 'filtered_no_infoRich':
            targeted_features_in_this_dataset = [value for value in dataframe_for_merging.columns.values if value not in combine_list_of_informoation_rich_features]
            select_data_for_merging = dataframe_for_merging[targeted_features_in_this_dataset]
        elif select == 'filtered_infoRich':
            targeted_features_in_this_dataset = [value for value in dataframe_for_merging.columns.values if value in combine_list_of_informoation_rich_features]
            select_data_for_merging = dataframe_for_merging[targeted_features_in_this_dataset]

        ## Don't need to select specific columns
        # 'pre_filter', 'filtered_excluded', 'filterd'
        else:
            select_data_for_merging = dataframe_for_merging

        ## Merging dataframes
        if i == 0:
            merged_dataframe = select_data_for_merging
        else:
            # similar to rbind() and keep all
            merged_dataframe = pandas.concat([merged_dataframe, select_data_for_merging], axis = 0, sort = True)

    # remove infoRich feature in raw_not_infoRich inputs
    # why it happens: feature_1 is not an information-rich feature in file_1 but is in file_2.
    #                 Merging raw_not_infoRich_file_1 and raw_not_infoRich_file_2 will include
    #                 feature_1. We need to remove it.
    if select == 'raw_not_infoRich':
        intersect_feature_name = list(set(info_rich_list).intersection(merged_dataframe.columns.values))
        merged_dataframe.drop(intersect_feature_name, axis = 1, inplace = True)

    EMMER_class.merged_dataframe = merged_dataframe.fillna(0)

    transform_info = Projection(merged_dataframe = EMMER_class.merged_dataframe, normalize = normalize)

    transform_info.V_df.to_csv(os.path.join(EMMER_class.output_dir, (select + "__transformation_matrix.csv")))
    transform_info.colmean_df.to_csv(os.path.join(EMMER_class.output_dir, (select + "__data_colmean.csv")))

    if normalize == True:
        transform_info.colstd_df.to_csv(os.path.join(EMMER_class.output_dir, (select + "__data_colstd.csv")))

    transform_info.projection_df.to_csv(os.path.join(EMMER_class.output_dir, (select + "__PCA_coordinates.csv")))
    transform_info.PC_annotation.to_csv(os.path.join(EMMER_class.output_dir, (select + "__percent_explained.csv")))

    notebook = UpdateNoteBook(notebook_name = notebook_name, neglect = neglect)
    notebook.updateMergeResult(merge_what = select, num_feature = merged_dataframe.shape[1], norm_eigen = transform_info.norm_eigvals)

    return(transform_info)


def sanityCheck(EMMER_class, input_file_names, current_filter, notebook_name, neglect, normalize, make_plot = False):
    """
    For generate datasets and plots for santiy check
    """
    if current_filter != 'None':
        Projection_class_in_list = [mergeDataFrame(EMMER_class = EMMER_class, select = 'pre_filter', file_name_list = emmer_result.pre_filter_data_file_names, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered_excluded', file_name_list = emmer_result.filter_out_data_file_names, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered', file_name_list = emmer_result.clean_df_file_names, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered_no_infoRich', file_name_list = emmer_result.clean_df_file_names, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered_infoRich', file_name_list = emmer_result.clean_df_file_names, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'raw_not_infoRich', file_name_list = emmer_result.raw_not_infoRich_data_name, normalize = normalize,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect)]
        input_file_in_list_of_list = [input_file_names]
        output_file_name_in_list = [(EMMER_class.output_file_tag + '__pre_filter_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_excluded_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_no_infoRich_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_infoRich_projection.csv'),
                                    (EMMER_class.output_file_tag + '__raw_not_infoRich_projection.csv')]
        index_tag_list = ['__pre_filter_projection.csv', '__filtered_excluded_projection.csv', '__filtered_projection.csv',
                          '__filtered_no_infoRich_projection.csv', '__filtered_infoRich_projection.csv', '__raw_not_infoRich_projection.csv']
        dataset_list = ['pre_filter', 'filtered_excluded', 'filtered', 'filtered_no_infoRich', 'filtered_infoRich', 'raw_not_infoRich']

        for p in range(6):
            Projection_class_in_list[p].projection_df.to_csv(output_file_name_in_list[p])

            if 'PC3' not in Projection_class_in_list[p].projection_df:   # TODO: should work, but need testing
                new_projection_df = Projection_class_in_list[p].projection_df.iloc[:, 0:2]
                original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:2])
                new_projection = numpy.array(new_projection_df)
            else:
                new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
                original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:3])

            notebook = UpdateNoteBook(notebook_name = notebook_name, neglect = neglect)

            if new_projection.size != 0:
                mtx1, mtx2, disparity = procrustes(original_projection, new_projection)
                notebook.updateProcrustesResult(procrustes_score = disparity, which_dataset = dataset_list[p], no_procrustes_score = False)
            else:
                notebook.updateProcrustesResult(procrustes_score = '', which_dataset = '', no_procrustes_score = True)


#        original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:3])
#
#        for p in range(6):
#            Projection_class_in_list[p].projection_df.to_csv(output_file_name_in_list[p])
#
#            new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
#            #if 'PC3' not in Projection_class_in_list[p].projection_df: TODO:
#            #    new_projection_df = Projection_class_in_list[p].projection_df.iloc[:, 0:2]
#            #    r, c = new_projection_df.shape
#            #    new_projection_df.iloc[3] = [0] * r
#            #    new_projection = numpy.array(new_projection_df)
#            #else:
#            #    new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
#            notebook = UpdateNoteBook(notebook_name = notebook_name, neglect = neglect)
#            if new_projection.size != 0:
#                mtx1, mtx2, disparity = procrustes(original_projection, new_projection)
#                notebook.updateProcrustesResult(procrustes_score = disparity, which_dataset = dataset_list[p], no_procrustes_score = False)
#            else:
#                notebook.updateProcrustesResult(procrustes_score = '', which_dataset = '', no_procrustes_score = True)


    else:
        Projection_class_in_list = [mergeDataFrame(EMMER_class = EMMER_class, select = 'pre_filter', file_name_list = emmer_result.pre_filter_data_file_names,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered_no_infoRich', file_name_list = emmer_result.clean_df_file_names,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect),
                                    mergeDataFrame(EMMER_class = EMMER_class, select = 'filtered_infoRich', file_name_list = emmer_result.clean_df_file_names,
                                                   info_rich_list = emmer_result.collections_of_info_rich_features, notebook_name = notebook_name, neglect = neglect)]
        input_file_in_list_of_list = [input_file_names]
        output_file_name_in_list = [(EMMER_class.output_file_tag + '__pre_filter_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_no_infoRich_projection.csv'),
                                    (EMMER_class.output_file_tag + '__filtered_infoRich_projection.csv')]
        index_tag_list = ['__pre_filter_projection.csv', '__filtered_no_infoRich_projection.csv', '__filtered_infoRich_projection.csv']
        dataset_list = ['pre_filter', 'filtered_no_infoRich', 'filtered_infoRich']
#        original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:3])
#
#        for p in range(3):
#            Projection_class_in_list[p].projection_df.to_csv(output_file_name_in_list[p])
#            new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
#
#            # FIXME
#            #if 'PC3' not in Projection_class_in_list[p].projection_df:
#            #    new_projection_df = Projection_class_in_list[p].projection_df.iloc[:, 0:2]
#            #    r, c = new_projection_df.shape
#            #    new_projection_df.iloc[3] = [0] * r
#            #    new_projection = numpy.array(new_projection_df)
#            #else:
#            #    new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
#            notebook = UpdateNoteBook(notebook_name = notebook_name, neglect = neglect)
#            if new_projection.size != 0:
#                mtx1, mtx2, disparity = procrustes(original_projection, new_projection)
#                notebook.updateProcrustesResult(procrustes_score = disparity, which_dataset = dataset_list[p], no_procrustes_score = False)
#            else:
#                notebook.updateProcrustesResult(procrustes_score = '', which_dataset = '', no_procrustes_score = True)


        # works, but need unittest
        for p in range(3):
            Projection_class_in_list[p].projection_df.to_csv(output_file_name_in_list[p])

            if 'PC3' not in Projection_class_in_list[p].projection_df:
                new_projection_df = Projection_class_in_list[p].projection_df.iloc[:, 0:2]
                original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:2])
                new_projection = numpy.array(new_projection_df)
            else:
                new_projection = numpy.array(Projection_class_in_list[p].projection_df.iloc[:, 0:3])
                original_projection = numpy.array(Projection_class_in_list[0].projection_df.iloc[:, 0:3])

            notebook = UpdateNoteBook(notebook_name = notebook_name, neglect = neglect)

            if new_projection.size != 0:
                mtx1, mtx2, disparity = procrustes(original_projection, new_projection)
                notebook.updateProcrustesResult(procrustes_score = disparity, which_dataset = dataset_list[p], no_procrustes_score = False)
            else:
                notebook.updateProcrustesResult(procrustes_score = '', which_dataset = '', no_procrustes_score = True)


    if make_plot == True:
        output_file_name_in_list = EMMER_class.output_dir + EMMER_class.output_file_tag + '__sanity_check_PCAs.pdf'
        sanity_check_plots = Plot(Projection_class_in_list, input_file_in_list_of_list, output_file_name_in_list, current_filter)
        sanity_check_plots.viewSanityCheckPlots()


##==2==## running emmer.harvest
if __name__ == '__main__':
    processed_args = HarvestArgs(suppress = False, silence = False, neglect = False)
    processed_args.processArgs()

    emmer_result = EMMER(input_dir = processed_args.input_dir, output_file_tag = processed_args.output_file_tag,
                         detection_limit = processed_args.detection_limit, tolerance = processed_args.tolerance,
                         filter = processed_args.filter, upper_threshold_factor = processed_args.upper_threshold_factor,
                         lower_threshold_factor = processed_args.lower_threshold_factor, num_cpu = processed_args.num_cpu,
                         specific_csv = processed_args.specific_csv, infoRich_threshold = processed_args.infoRich_threshold,
                         notebook_name = processed_args.notebook_name, neglect =  processed_args.neglect,
                         quick_look = processed_args.quick_look, normalize = processed_args.normalize,
                         use_fractional_abundance = processed_args.use_fractional_abundance)

    if processed_args.specific_csv == True:
        emmer_result.singleFile()
    else:
        if len(processed_args.input.input_files) > 1:
            emmer_result.multipleFiles()
            transform_info = mergeDataFrame(EMMER_class = emmer_result, select = 'filtered_infoRich',
                                            file_name_list = emmer_result.clean_df_file_names,
                                            info_rich_list = emmer_result.collections_of_info_rich_features,
                                            notebook_name = processed_args.notebook_name, normalize = processed_args.normalize,
                                            neglect =  processed_args.neglect)
        else:
            emmer_result.singleFile()

    print("computation time: %s seconds " % (time.time() - start_time))
    run_time = (time.time() - start_time)

    if processed_args.plot_result == True and processed_args.sanity_check_plots == False:
        processed_args.output_file_name = emmer_result.output_dir + emmer_result.output_file_tag + '__PCA_w_info_rich.pdf'
        visualizing_result = Plot(Projection_class_in_list = [transform_info],
                                  input_file_in_list_of_list = [processed_args.input.input_files],
                                  output_file_name = processed_args.output_file_name,
                                  filter = processed_args.filter)
        visualizing_result.viewSinglePCAplot()

    if processed_args.sanity_check_plots == True:
        os.chdir(emmer_result.output_dir)
        sanityCheck(EMMER_class = emmer_result, input_file_names = processed_args.input.input_files,
                    current_filter = processed_args.filter, make_plot = processed_args.plot_result,
                    notebook_name = processed_args.notebook_name, neglect =  processed_args.neglect,
                    normalize = processed_args.normalize)

    notebook = UpdateNoteBook(notebook_name = processed_args.notebook_name, neglect =  processed_args.neglect).updateRunTime(run_time = run_time)

    if processed_args.neglect == False:
        print('\nAnalytical parameters and brief summary stores at:\n' + processed_args.notebook_name)

    if os.cpu_count() > processed_args.args.c and processed_args.args.c == 1:
        print(f'\n[!] Tip: your computer has {os.cpu_count()} CPUs, but you only used one to run emmer.\n    Please consider to speed up your analysis by assign more than one CPU when running emmer with -c option.\n    Suggested number of CPU: {os.cpu_count() - 1}\n')



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_individual

from ..bake import BakeCommonArgs
from ..main.basic.read import RetrospectDataImport, GetFiles
from ..posthoc.visual.individual import IndividualArgs
from ..troubleshoot.err.error import Error, ErrorCode8

import unittest
import sys
import os


class TestIndividualArgs(unittest.TestCase):

    def test_IndividualArgs(self):
        print('\ntest_IndividualArgs:')
        print('        case 1: error handling')
        print('             1.1: missing args.i')
        sys.argv[1:] = ['-m', 'Individual']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode8):
            individual_args = IndividualArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True, silence = False)


        print('        ---------------------------------------------------')
        print('             1.2: giving both args.i and args.p')
        sys.argv[1:] = ['-m', 'Individual', '-p', 'emmer/data/bake_data_dir_7/_retrospect_individaul_coloring_parameter.csv',
                        '-i', 'emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        #with self.assertRaises(WarningCode13):
        individual_args = IndividualArgs(args = processed_args.args, current_wd = current_wd,
                                         suppress = True, silence = False)
        my_result = individual_args.warning_code
        expected_result = '13'
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             1.3: args.i is not a csv file')
        sys.argv[1:] = ['-m', 'Individual', '-i', 'emmer/data/problem_maker/read_module/not_a_csv_file.txt']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode8):
            individual_args = IndividualArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True, silence = False)


        print('        ---------------------------------------------------')
        print('        case 2: expect to work')
        sys.argv[1:] = ['-m', 'Individual', '-p', 'emmer/data/bake_data_dir_7/_retrospect_individaul_coloring_parameter.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        individual_args = IndividualArgs(args = processed_args.args, current_wd = current_wd,
                                         suppress = True, silence = False)
        print(type(individual_args))
        print(individual_args.pca_coordinate.coordinate_w_info.shape)
        print('===========================================================')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
   name='piemmer',
   version='1.0.4-dev',
   description='A algorithm to simplify the input for principal component analysis',
   author='Hao-Wei chang',
   author_email='emmer.man42@gmail.com',
   packages=find_packages(),
   package_data={'piemmer': ['data/*.csv', 'data/*/*.csv', 'data/*/*/*.csv']},
   install_requires=['numpy', 'pandas', 'matplotlib', 'scikit-bio', 'scipy', 'tqdm', 'statsmodels'],
)



#!/usr/bin/env python3

from ..warn.warning import *
from ..err.error import *

import matplotlib.colors as mcolors
from matplotlib import markers
import sys
import re



"""
Handle all the input messages in EMMER
"""

## TODO: no second change after failing input evaluation
#        python3 -m emmer.bake
#        x


class EvaluateInput:
    """
    Evaluate input from users.

    Arguments:
        input -- Type: list
                 User input
        set -- Type: set
               Set of corresponding elements that require user-input arguments
        suppress -- Type: boolean
                    Exit program after raise error or not
        second_chance -- Type: boolean
                         Give user a second chance to change the input argument or not
        expect -- Type: list or 0
                  Expected input values in list

    Attributes:
        input -- Type: list
        set -- Type: list
        suppress -- Type: boolean
        expect -- Type: list or 0
        map_input -- Type: dictionary
                     key: elements in set; item: user input
        passed -- Type: boolean
                  True: pass the evaluation; False: does not pass the evaluation but
                  users still have a second change to input arguments
    """

    def __init__(self, input, set, suppress, expect = 0):
        self.input = input
        self.set = list(set)
        self.suppress = suppress
        self.expect = expect
        self.passed = False


        print('self.input')
        print(self.input)
        print('self.set')
        print(self.set)

        ## the number of elements in input are as expected
        try:
            if len(set) != len(input):
                raise Error(code = '10')
            else:
                self.map_input = dict(zip(set, input))
                self.passed = True
        except Error as e:
            raise ErrorCode10(suppress = suppress)


    def checkDeveloperError(self):
        try:
            if self.expect != 0:
                raise Error(code = '31')
        except Error as e:
            raise ErrorCode31(suppress = self.suppress) from e


    def evaluateColor(self):
        """
        InputCode1 specific evaluation
        """
        self.passed = False
        EvaluateInput.checkDeveloperError(self)

        for element in self.input:
            if element in mcolors.cnames.keys() or element == 'none':
                # allow element == 'none' will give user to plotting partial data
                # onto the PCA plot
                self.passed = True
            else:
                hex_color = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', element)
                # ref: https://stackoverflow.com/questions/30241375/python-how-to-check-if-string-is-a-hex-color-code

                try:
                    if not hex_color:
                        raise Error(code = '28')   ## FIXME: unable to arise this error
                    else:
                        self.passed = True
                except Error as e:
                    raise ErrorCode28(suppress = self.suppress) from e


    def evaluateStatCluster(self):
        """
        InputCode2 specific evaluation
        """
        self.passed = False
        EvaluateInput.checkDeveloperError(self)

        # expect to have more than two clusters when running statistical test
        try:
            if len(list(set(self.input))) < 2:
                raise Error(code = '37')
            else:
                self.passed = True
        except Error as e:
            raise ErrorCode37(suppress = self.suppress) from e


    def evaluateModeOption(self):
        """
        InputCode3 specific evaluation

        ErrorCode9 is related to WarningCode9. When WarningCode9 arised, user can
        try to advert the error by providing appropriate input via InputCode3. However,
        if user fail to provide expected input, ErrorCode9 will arise and terminate the
        program. No second chance for re-entering.
        """
        self.passed = False

        try:
            if self.expect == 0:
                raise Error(code = '7')
            elif len(set(self.input).intersection(list(self.expect.values()) + list(self.expect.keys()))) == 0:
                raise Error(code = '9')
            else:
                self.passed = True
        except Error as e:
            if e.code == '7':
                raise ErrorCode7(suppress = self.suppress) from e
            elif e.code == '9':
                raise ErrorCode9(suppress = self.suppress) from e


    def evaluateMarker(self):
        """
        InputCode4 specific evaluation
        """
        self.passed = False
        EvaluateInput.checkDeveloperError(self)

        try:
            for element in self.input:
                if element not in markers.MarkerStyle.markers.keys():
                    raise Error(code = '33')
                else:
                    self.passed = True
        except Error as e:
            raise ErrorCode33(suppress = self.suppress) from e


    def evaluateGroupOption(self):
        """
        InputCode5 specific evaluation
        """
        self.passed = False

        try:
            if self.expect == 0:
                raise Error(code = '7')
            elif len(set(self.input).intersection(list(self.expect.values()) + list(self.expect.keys()))) == 0:
                raise Error(code = '36')
            else:
                self.passed = True
        except Error as e:
            if e.code == '7':
                raise ErrorCode7(suppress = self.suppress) from e
            elif e.code == '36':
                raise ErrorCode36(suppress = self.suppress) from e

## TODO: unclear whether I am assign to edge or fill
class InputCode1:
    """
    Input edge or fill color:
    Please assign color to each group or individuals.
    Please use semicolon to separate the assignment for each group or individuals.

    Arguments:
        set -- Type: list
        suppress -- Type: boolean

    Attributes:
        color_dict -- Type: dictionary
                      key: group; item: color
    """

    def __init__(self, set, suppress):
        """
        [[Input code 1]]
        Input color information:
        Please assign color to each group or individuals.
        Please use semicolon to separate the assignment for each group or individuals.

        Example:
        Assuming you have two groups, group_A and group_B. You want to assign red and blue to
        group_A and group_B respectively. You can enter:
        red;blue
        #ff0000;#0000ff
        #FF0000;#0000FF
        """
        print(self.__init__.__doc__)

        ## take input
        option = input('\n        Please enter your input:\n').split(';')

        ## evaluate input
        confirmed_input = EvaluateInput(input = option, set = set, suppress = True)

        if confirmed_input.passed == False:
            option = input('\n        Please enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        confirmed_input.evaluateColor()
        if confirmed_input.passed == False:
            option = input('\n        Please renter your input:\n').split(';')
            confirmed_input = confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        ## export input
        self.color_dict = confirmed_input.map_input


class InputCode2:
    """
    Input cluster assignment for PERMANOVA test.

    Arguments:
        set -- Type: list
        choice_dict -- Type: dictionary
                       Key: option number; item: full name of the option
        suppress -- Type: boolean

    Attributes:
        cluster_dict -- Type: dictionary
                        key: group; item: cluster
    """

    def __init__(self, set, suppress):
        """
        [[Input code 2]]
        Input cluster assignment:
        Please assign cluster to each group for PERMANOVA test.
        """
        print(self.__init__.__doc__)

        ## take input
        option = input('\n        Please re-enter your input:\n').split(';')

        ## evaluate input
        confirmed_input = EvaluateInput(input = option, set = set, suppress = True)

        if confirmed_input.passed == False:
            option = input('\n        Please re-enter your input:').split(';')
            confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        confirmed_input.evaluateStatCluster()
        if confirmed_input.passed == False:
            option = input('\n        Please re-enter your input:').split(';')
            confirmed_input = confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        ## export input
        self.cluster_dict = confirmed_input.map_input


class InputCode3:
    """
    Choose emmer.bake mode.

    In the future, this input class can be generalized to all of the multiple choice
    input in emmer

    Arguments:
        set -- Type: list
        choice_dict -- Type: dictionary
                       Key: option number; item: full name of the option
        suppress -- Type: boolean

    Attributes:
        selection -- Type: str
                     The full name of the option chose by the user
    """

    def __init__(self, set, choice_dict, suppress):
        """
        [[Input code 3]]
        Please choose emmer.bake mode. You can choose from the following options:

        [1] Individual
        [2] Permanova
        [3] RevisitThreshold
        [4] Reproducibility
        [5] Bifurication
        [6] Projection

        After you make your choose, please type the full name of the mode (e.g.
        'Individual') or option number (e.g. '1'). Please note that emmer.bake
        are case sensitive. In another word, 'individual' will not be accepted
        as legitimate input.

        Additional explanation on those methods:
        1. Individual: plot individual samples onto the PCA plot. Allow better flexibility in
                       aesthetics setting
        2. Permanova: conduct PERMANOVA test on different groups in the PCA plot
        3. RevisitThreshold: revisit threshold settings (args.t, args.l, and args.u) used in
                             emmer.harvest
        4. Reproducibility: summarizing information-rich feature calling reproducibility into
                            basic statistics and histogram
        5. Bifurication: identify information-rich features that help to differentiate different
                         groups on the PCA space.
        6. Projection: project new observation onto an existing PCA space.

        Please make your choice:
        """
        print(self.__init__.__doc__)

        ## take input
        option = input('\n        Please re-enter your input:\n').split(';')

        ## evaluate input
        confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = choice_dict, suppress = True)

        if confirmed_input.passed == False:
            option = input('\n        Please re-enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = choice_dict, suppress = False)

        confirmed_input.evaluateModeOption()
        if confirmed_input.passed == False:
            option = input('\n        Please re-enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = choice_dict, suppress = False)

        ## export input
        if option[0] in choice_dict.keys():
            self.selection = choice_dict[option[0]]
        else:
            self.selection = option[0]


class InputCode4:
    """
    Please assign shape to each group or individuals.

    Arguments:
        set -- Type: list
        choice_dict -- Type: dictionary
                       Key: option number; item: full name of the option
        suppress -- Type: boolean

    Attributes:
        shape_dict -- Type: dictionary
                      key: group; item: shape
    """

    def __init__(self, set, suppress):
        """
        [[Input code 4]]
        Please assign shape to each group or individuals.
        Please semicolon comma to separate the assignment for each group or individuals.
        More shape information (https://matplotlib.org/api/markers_api.html).
        """
        print(self.__init__.__doc__)

        ## take input
        option = input('\n        Please enter your input:\n').split(';')

        ## evaluate input
        confirmed_input = EvaluateInput(input = option, set = set, suppress = True)

        if confirmed_input.passed == False:
            option = input('\n        Please enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        confirmed_input.evaluateMarker()
        if confirmed_input.passed == False:
            option = input('\n        Please enter your input:\n').split(';')
            confirmed_input = confirmed_input = EvaluateInput(input = option, set = set, suppress = False)

        ## export input
        self.shape_dict = confirmed_input.map_input


class InputCode5:
    """
    Allow users to choose whether they want to color by group or by individuals

    Arguments:
        set -- Type: list
        choice_dict -- Type: dictionary
                       Key: option number; item: full name of the option
        suppress -- Type: boolean

    Attributes:
        decision -- Type: string
    """

    def __init__(self, suppress):
        """
        [[Input code 5]]
        Please choose whether you want to color by group or by individuals.
        You can choose from the following options:

        [1] Group
        [2] Individaul

        Please make your choice:
        """
        print(self.__init__.__doc__)

        ## take input
        option = input('\n        Please enter your input:\n').split(';')
        decision_dict = {'1': 'Group', '2': 'Individaul'}

        ## evaluate input
        confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = decision_dict, suppress = True)

        if confirmed_input.passed == False:
            option = input('\n        Please enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = decision_dict, suppress = False)

        confirmed_input.evaluateGroupOption()
        if confirmed_input.passed == False:
            option = input('\n        Please enter your input:\n').split(';')
            confirmed_input = EvaluateInput(input = option, set = ['mode'], expect = decision_dict, suppress = False)

        ## export input
        self.decision = option


class InputCode6:
    """
    Allow users to add notes to the emmer_notebook

    Arguments:
        suppress -- Type: boolean

    Attributes:
        input_string -- Type: string
    """

    def __init__(self, suppress):
        """
        [[Input code 6]]
        Please enter an additional note in the emmer_notebook. Hit Enter once
        you finish entering your note.

        Your entry:
        """
        print(self.__init__.__doc__)

        self.input_string = input()


class InputCode7:
    """
    Reminding users that running emmer.tests will remove existing output files.

    Arguments:
        suppress -- Type: boolean

    Attributes:
        input_string -- Type: string
    """

    def __init__(self, suppress):
        """
        [[Input code 7]]
        Please note that running emmer.tests will remove existing output files.
        Please remember to backup those output files if you still want to use
        them in the future.

        Please enter "yes" when you ready to proceed?
        """
        print(self.__init__.__doc__)

        answer = input()

        if answer != "yes":
            sys.exit()



## usage
# at a level above emmer/
# python3 -m emmer.test.test_bifurication

from ..bake import BakeCommonArgs
from ..posthoc.stats.bifurication import BifuricationArgs, linearRegressionPVal, DifferentiatingFeatures
from ..posthoc.visual.viewer import Projection
from ..troubleshoot.err.error import ErrorCode12, ErrorCode21

import numpy.testing
import unittest
import pandas
import numpy
import sys
import os


class TestBifuricationArgs(unittest.TestCase):

    def test_BifuricationArgs(self):
        print('\ntest_BifuricationArgs.:')
        print('        case 1: error handling')
        print('             1.1: missing both args.p and args.i setting. raise error for missing args.p')
        sys.argv[1:] = ['-m', 'Bifurication']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode12):
            bifurication_args = BifuricationArgs(args = processed_args.args, current_wd = current_wd,
            	                                 suppress = True)

        print('        ---------------------------------------------------')
        print('             1.2: missing args.i setting')
        sys.argv[1:] = ['-m', 'Bifurication', '-p', 'emmer/data/bake_data_dir_4/information_rich_features_summary.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode21):
            bifurication_args = BifuricationArgs(args = processed_args.args, current_wd = current_wd,
                                                 suppress = True)

        print('        ---------------------------------------------------')
        print('             1.3: args.1 only contains one csv file')
        sys.argv[1:] = ['-m', 'Bifurication', '-p', 'emmer/data/bake_data_dir_4/information_rich_features_summary.csv',
                        '-i', 'emmer/data/data_dir_1']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode21):
            bifurication_args = BifuricationArgs(args = processed_args.args, current_wd = current_wd,
                                                 suppress = True)

        print('        ---------------------------------------------------')
        print('        case 2: expect to work')
        print('             2.1: get correct list of information-rich features')
        sys.argv[1:] = ['-m', 'Bifurication', '-p', 'emmer/data/bake_data_dir_4/information_rich_features_summary.csv',
                        '-i', 'emmer/data/bake_data_dir_4/filtered_data']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        bifurication_args = BifuricationArgs(args = processed_args.args, current_wd = current_wd,
                                             suppress = True)

        my_result = bifurication_args.list_of_info_rich
        expected_result = ['ASV_1', 'ASV_15', 'ASV_2', 'ASV_3', 'ASV_32', 'ASV_4', 'ASV_7']

        self.assertListEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             2.2: get correct number of input files')
        my_result = len(bifurication_args.input_files)
        expected_result = 2

        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestLinearRegressionPVal(unittest.TestCase):

    def test_linearRegressionPVal(self):
        print('\ntest_linearRegressionPVal:')
        A = [[1,15,2], [1,11,1], [1,13,1], [0,2,1], [0,1,2], [0,3,1]]
        A_df = pandas.DataFrame(A, columns=['y','x1','x2'], index=['A__s1','A__s2','A__s3','B__s4','B__s5','B__s6'])
        target = A_df['y']
        data = A_df[['x1','x2']]

        my_result = numpy.array(linearRegressionPVal(target = target, data = data, silence_intersect = False))
        expected_result = numpy.array([0.7805, 0.0046, 0.6640]).reshape(3, 1)
            # result generated in R use lm() function

        numpy.testing.assert_almost_equal(my_result, expected_result, decimal = 4)
        print('===========================================================')


class TestDifferentiatingFeatures(unittest.TestCase):

    def test_differentiatingFeatures(self):
        print('\ntest_differentiatingFeatures:')
        print('        Case 1: dataset has a differentiating feature')
        A = [[15,2], [11,1], [13,1], [2,1], [1,2], [3,1]]
        A_df = pandas.DataFrame(A, columns=['feature_1','feature_2'], index=['A__s1','A__s2','A__s3','B__s4','B__s5','B__s6'])

        Projection_class_object = Projection(merged_dataframe = A_df, normalize = False)
        Projection_class_object.cleanSpec()

        DifferentiatingFeatures_class_object = DifferentiatingFeatures(Projection_class_object)
        DifferentiatingFeatures_class_object.atGroup()

        my_result = DifferentiatingFeatures_class_object.differentiating_feature

        expected_result = ['feature_1']

        self.assertListEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        Case 2: dataset do not have differentiating feature')
        C = [[1,2], [3,1], [2,1], [2,1], [1,2], [3,1]]
        C_df = pandas.DataFrame(C, columns=['x1','x2'], index=['A__s1','A__s2','A__s3','B__s4','B__s5','B__s6'])

        Projection_class_object = Projection(merged_dataframe = C_df, normalize = False)
        Projection_class_object.cleanSpec()

        DifferentiatingFeatures_class_object = DifferentiatingFeatures(Projection_class_object)
        DifferentiatingFeatures_class_object.atGroup()

        my_result = len(DifferentiatingFeatures_class_object.differentiating_feature)

        expected_result = 0
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_bake

from ..main.advanced.iteration import MinusOneVNE, InfoRichCalling, reproducibility_summary
from ..main.basic.read import RawDataImport, RetrospectDataImport, GetFiles
from ..bake import BakeCommonArgs

import argparse
import unittest
import sys


class TestBakeCommonArgs(unittest.TestCase):

    def test_getArgsM(self):
        print('\ntest_TestBakeCommonArgs.getArgsM:')
        print('        case 1: error handling')
        print('             1.1: missing args.m setting')
        sys.argv[1:] = []
        #with self.assertRaises(ErrorCode9):
        #    processed_args = BakeCommonArgs(suppress = True)
        #    processed_args.getArgsM()
        processed_args = BakeCommonArgs(suppress = False, test = True, neglect = True, silence = False)
        processed_args.getArgsM()
        my_result = processed_args.warning_code
        expected_result = '9'
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: unexpected args.m setting')
        sys.argv[1:] = ['-m', 'Unknown_mode']
        #with self.assertRaises(ErrorCode9):
        #    processed_args = BakeCommonArgs(suppress = True)
        #    processed_args.getArgsM()
        processed_args = BakeCommonArgs(suppress = False, test = True, neglect = True, silence = False)
        processed_args.getArgsM()
        my_result = processed_args.warning_code
        expected_result = '9'
        self.assertEqual(my_result, expected_result)

        print('===========================================================')


    def test_getHomeKeepingArgs(self):
        print('\ntest_TestBakeCommonArgs.getHomeKeepingArgs:')
        print('        case 1: get arguments')
        sys.argv[1:] = ['-m', 'Reproducibility']
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        my_result = processed_args.selected_model
        expected_result = 'Reproducibility'

        self.assertEqual(my_result, expected_result)
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Usage:
# move to one level above emmer/
# python3 -m emmer.tests

import unittest
from .troubleshoot.warn.warning import WarningCode6
from .troubleshoot.inquire.input import InputCode7

"""
Initiate all tests
"""


if __name__ == '__main__':
    # try:
    #     if path exist
    #         raise WarningCode6(silence = False)
    # except:
    #     inputCodeX

    answer = InputCode7(suppress = False)

    testsuite = unittest.TestLoader().discover('.')
    unittest.TextTestRunner(verbosity=1).run(testsuite)



#!/usr/bin/env python3

from ...troubleshoot.err.error import Error, ErrorCode41, ErrorCode46

import pandas
import numpy
import math

"""
This module stores the mathematical tools for converting the input matrix into
a density matrix and calculate the von Neumann entropy.
"""

class NonDesityMatrix:
    """
    Transform a numeric non-density matrix and calculate the von Neumann entropy.

    Arguments:
        data -- Type: RawDataImport.data (pandas.DataFrame)
        epsilon -- Type: int
                   Define machine percision. When set at '8', it means that the
                   machine percision is 10^-8
        use_cor_matrix -- Type: boolean
                          When set as True, calculated the eigenvalue from the
                          correlation matrix

    Attribute:
        data -- Type: numpy.ndarray
                Non-density matrix
        epsilon -- Type: float
                   Machine percision
        nrow -- Type: int
                Number of samples
        ncol -- Type: int
                Number of features
        colmean -- Type: numpy.ndarray
                   A matrix that has one row and <self.ncol> columns, each element in
                   this matrix represents the mean of its corresponding column in
                   <self.data>
        colstd -- Type: numpy.ndarray
                  A matrix that has one row and <self.ncol> columns, each element in
                  this matrix represents the standard deviation of its corresponding
                  column in <self.data>
        matrixForSVD -- Type: numpy.ndarray
                        A matrix that is transform from <self.data> and ready for SVD.
                        The transformation is conducted so the <self.matrixForSVD> will
                        satistify the following relationship:
                        eigendecompsition of <self.data> will generate eignevalue lamda
                        SVD of <self.matrixForSVD> will generate signular_value
                        lamda = (signular_value)^2
        normEigvals -- Type: numpy.ndarray
                       Normalized eigenvalues. The sum of all eigenvalues equals to one.
        vNE -- Type: numpy.float64
               The von Neumann entropy of the input numeric non-density matrix
        numZeroEig -- Type: numpy.int64
                      The number of normalized eigenvalues that equals to zero or less
                      than <self.epsilon>
    """
    def __init__(self, data, normalize, epsilon = 8, suppress = False):
        self.data = numpy.array(data)
        self.epsilon = 10 ** (-epsilon)
        self.normalize = normalize   # when True: use correlation matrix instead of covariance matrix
        self.nrow = numpy.size(self.data, 0)
        self.ncol = numpy.size(self.data, 1)
        self.suppress = suppress

        try:
            if self.nrow < 2:
                raise Error(code = '41')
        except Error as e:
            raise ErrorCode41(suppress = self.suppress) from e


    def normEigvals(self):
        """
        Prepare normalized eigenvalues for the von Neumann entropy calculation

        The math behide this computational process is:
        (signular_val of M)^2 = eigvals of (M * M.transpose)
        """

        ##--1--## preparing matrix
        # mean center the matrix:
        # the element in each column minus the corresponding colmean
        try:
            self.colmean = numpy.mean(numpy.array(self.data), axis = 0)
        except TypeError as e:
            raise ErrorCode46(suppress = self.suppress)

        S = numpy.tile(self.colmean, (self.nrow, 1))
        meanCenteredData = self.data - S

        if self.normalize == False:
            # equivalence of calculating eignevalues of a covariance matrix
            self.matrixForSVD = meanCenteredData
        else:
            # equivalence of calculating eignevalues of a correlation matrix
            self.colstd = numpy.std(numpy.array(meanCenteredData), ddof = 1, axis = 0)
            self.matrixForSVD = numpy.matmul(meanCenteredData, numpy.diag(1/self.colstd))

        ##--2--## calculate singular values
        s = numpy.linalg.svd(self.matrixForSVD, compute_uv = False)
        eigvalsFromSVD = numpy.power(s, 2)
        eigvalsFromSVD[eigvalsFromSVD < self.epsilon] = 0

        ##--3--## eigenvalues normalization
        normEigvals = eigvalsFromSVD/sum(eigvalsFromSVD)
        normEigvals[normEigvals < self.epsilon] = 0
        self.normEigvals = normEigvals
        return(self.normEigvals)


    def vNE(self):
        """
        Calculate the von Neumann entropy
        """
        beta = self.normEigvals()
        beta = numpy.array([i for i in beta if i != 0])
        self.vNE = sum(-numpy.transpose(beta)*numpy.log2(beta))
        return(self.vNE)


    def numZeroEig(self):
        """
        Report the number of eigenvalues that equals to zero in the vector of normalized eigenvalues
        """
        self.numZeroEig = sum(self.normEigvals() == 0)

        # the number of reported singular values equals to the min(self.nrow, self.ncol)
        if (self.nrow < self.ncol and self.numZeroEig > 0):
             self.numZeroEig = self.numZeroEig + (self.ncol - self.nrow)

        return(self.numZeroEig)



#!/usr/bin/env python3

from ..troubleshoot.err.error import Error, ErrorCode23
#from sklearn.linear_model import LinearRegression
from scipy import stats
import pandas
import numpy
import sys
import os
import io


"""
A tool box of functions that could be used in almost all of the scripts in emmer package.
"""

##==0==##
def flattern(list_of_list):  ## TODO: it works, but need test
    """
    Flattern list of list to list.

    Example:
    A = [[1, 2], [3, 4]]
    B = flattern(A)
    > B = [1, 2, 3, 4]
    """
    flattern_list = [item for sublist in list_of_list for item in sublist]
    return(flattern_list)


def emptyNumpyArray(nrow, ncol):
    """
    Create empty numpy array and fill it with NaN.
    """
    numpy_array = numpy.zeros([nrow, ncol])
    numpy_array[:] = numpy.nan
    return(numpy_array)


def toFloat(number_in_str, suppress = False):
    """
    Convert interger to float
    """
    try:
        num = float(number_in_str)
        return(num)
    except ValueError as e:
        raise ErrorCode23(suppress = suppress) from e


def floatRange(input_tuple):
    """
    Based on the input tuple, create a list of float that starts with input_tuple[1] and end with input_tuple[0] with
    the increment of input_tuple[2]

    Note: cannot use range because it only accept int
    """
    max = input_tuple[0]
    min = input_tuple[1]
    step = input_tuple[2]

    float_list = []

    if max == min:
        float_list = [min]

    while min < (max + step):
        float_list.append(min)
        min += step

    return(float_list)


def addElementsInList(list_1, list_2):
    """
    list_1 = [A, B, C]
    list_2 = [X, Y, Z]

    list_3 = addElementsInList(list_1, list_2)
    > list_3 = [A+X, B+Y, C+Z]
    """
    ## TODO: current function is valnerable to error cause be uneven list length
    result_list = []
    for i in range(0, len(list_1)):
        result_list.append(list_1[i] + list_2[i])
    return(result_list)


def dualAssignment(dataframe, sep):
    """
    if columns == 'A-B' and sep = '-'
        separated this column into two columns

        if column 'A' already exist, but not column 'B':
            column 'A' + column 'A-B'
            create column 'B'
    """
    cols_need_works = [col for col in dataframe.columns if sep in col]
    dataframe_need_work = dataframe[cols_need_works]
    cols_not_need_works = [col for col in dataframe.columns if sep not in col]
    dataframe_not_need_work = dataframe[cols_not_need_works]

    dict = {}
    for c_1 in range(dataframe_not_need_work.shape[1]):
        list_to_hash = dataframe_not_need_work.iloc[:, c_1]
        key = dataframe_not_need_work.columns[c_1]
        if key in dict.keys():
            list_1 = dict[key]
            dict[key] = addElementsInList(list_1, list_to_hash)
        else:
            dict[key] = list_to_hash

    for c_2 in range(dataframe_need_work.shape[1]):
        list_2 = list(dataframe_need_work.iloc[:, c_2])
        key_list = str(dataframe_need_work.columns[c_2]).split(sep)

        for k in range(len(key_list)):
            if key_list[k] in dict.keys():
                list_1 = dict[key_list[k]]
                dict[key_list[k]] = addElementsInList(list_1, list_2)
            else:
                dict[key_list[k]] = list_2

    return(dict)



#!/usr/bin/env python3


from ...main.basic.read import GetFiles, RetrospectDataImport
from ...troubleshoot.err.error import *
from ...troubleshoot.warn.warning import *

import matplotlib.pyplot as plt
import pandas
import numpy
import sys
import os

"""
Support emmer.bake Reproducibility mode

Summarizing information-rich feature calling reproducibility into basic statistics
and histogram
"""

class ReproducibilitySummary:
    """
    Summarizing information-rich feature calling statistics
    """
    def __init__(self, EMMER_reproducibility_df):
        reproducibility_in_list_of_list = EMMER_reproducibility_df.values.tolist()
        self.reproducibility = numpy.array([item for sublist in reproducibility_in_list_of_list for item in sublist])
        self.reproducibility_no_zero = self.reproducibility[self.reproducibility != 0]
        self.median = numpy.median(self.reproducibility_no_zero)
        self.mean = numpy.mean(self.reproducibility_no_zero)
        self.sd = numpy.std(self.reproducibility_no_zero)


class ReproducibilityArgs:
    """
    Take common arguments for Reproducibility modes when running bake modules

    Objective: so we can test and use @

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        current_wd -- Type: str
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean

    Attributes:
        args -- Type: argparse.Namespace
        current_wd -- Type: str
        suppress -- Type: boolean
        silence -- Type: boolean
        bin_num -- Type: int
                   Number of bin used when generating histogram
        reproducibility_list -- Type: numpy.ndarray
                                One row; mulitple columns. A list of non-zero information-rich
                                feature calling reproducibility values
        input_file_num -- Type: interger
        reproducibility_summary_df -- Type: pandas.DataFrame
        warning_code -- Type: str
                        for unittest

    """
    def __init__(self, args, current_wd, suppress, silence):
        self.args = args
        self.current_wd = current_wd
        self.suppress = suppress
        self.silence = silence

        try:
            if self.args.b:
                self.bin_num = self.args.b
            else:
                raise WarningCode7(silence = self.silence)
        except WarningCode7:
            self.warning_code = '7'
            self.bin_num = 20

        try:
            if not self.args.i:
                raise Error(code = '1')
        except Error as e:
            raise ErrorCode1(suppress = self.suppress)

        if os.path.isdir(os.path.join(self.current_wd, self.args.i)) == False:
            # works on single csv file
            repro = RetrospectDataImport(file_name = os.path.join(self.current_wd, self.args.i), type = 'reproducibility')  # TODO: FileNotFoundError
            repro_summary = ReproducibilitySummary(repro.reproducibility)
            print(f'Reproducibility summary:\nmean {repro_summary.mean}\nmedian: {repro_summary.median}\nstandard deviation: {repro_summary.sd}\n')
            self.reproducibility_list = repro_summary.reproducibility_no_zero

        else:
            repro_files = GetFiles(self.args.i)
            repro_files_list = repro_files.input_files
            reproducibility_list = []

            for f in range(len(repro_files_list)):
                repro_sub = RetrospectDataImport(file_name = os.path.join(self.current_wd, repro_files_list[f]), type = 'reproducibility')
                repro_summary_sub = ReproducibilitySummary(repro_sub.reproducibility)
                reproducibility_list_sub = list(repro_summary_sub.reproducibility_no_zero)

                if f == 0:
                    reproducibility_list = reproducibility_list_sub
                    reproducibility_summary_df = repro_sub.reproducibility
                else:
                    reproducibility_list = reproducibility_list + reproducibility_list_sub
                    reproducibility_summary_df = reproducibility_summary_df.join(repro_sub.reproducibility, how = 'outer')  # similar to R merge(df.1, df.2, by = ...)

            self.input_file_num = len(repro_files_list)
            self.reproducibility_summary_df = reproducibility_summary_df.fillna(0)
            self.reproducibility_list = numpy.array(reproducibility_list)
            print(f'Reproducibility summary:\nmean {numpy.mean(self.reproducibility_list)}\nmedian: {numpy.median(self.reproducibility_list)}\nstandard deviation: {numpy.std(self.reproducibility_list)}\n')


def reproSummary(args, current_wd, retrospect_dir, output_file_tag, suppress, silence):
    #    # python3 -m emmer.bake -m 'Reproducibility' -b 20 -i emmer/data/bake_data_dir_4/information_rich_features_summary.csv

        ## take-in args
        reproducibility_args = ReproducibilityArgs(args = args, current_wd = current_wd, suppress = False, silence = False)

        if(reproducibility_args.input_file_num > 1):
            output_df_file_name = os.path.join(retrospect_dir, (output_file_tag + 'reproducibility_summary.csv'))
            reproducibility_args.reproducibility_summary_df.to_csv(output_df_file_name)

        ## make histogram
        fig = plt.figure()
        n, bins, patches = plt.hist(reproducibility_args.reproducibility_list, reproducibility_args.bin_num, range = [0, 100], facecolor='black', alpha=0.5)
        plt.xticks(numpy.arange(0, 110, 10))
        plt.xlabel('Reproducibility (%)')
        plt.ylabel('Counts')
        plt.show()
        output_file_name = os.path.join(retrospect_dir, (output_file_tag + 'reproducibility_hist.pdf'))
        fig.savefig(output_file_name, idp = 300)
